# encoding: UTF-8
class Account < ActiveRecord::Base
  has_paper_trail
  include Uuidextension
  include ApplicationHelper
  has_many :outgoing_transfers, :class_name => 'Transfer',
           :finder_sql => proc {"
SELECT transfers.*, 0 as reversed FROM transfers 
  JOIN account_relationships  ON  (transfers.id_account_from = account_relationships.child_id)
    WHERE account_relationships.parent_id = \'#{id}\' #{additional_where_sql}
    ORDER BY datum, seq" }

  has_many :incoming_transfers, :class_name => "Transfer",
           :finder_sql => proc {"
SELECT transfers.*, 1 as reversed FROM transfers 
  JOIN account_relationships  ON  (transfers.id_account_to = account_relationships.child_id)
    WHERE account_relationships.parent_id = \'#{id}\' #{additional_where_sql}
    ORDER BY datum, seq"}

# envelopes ordering
  has_many :all_transfers, :class_name => "Transfer",
           :finder_sql => proc { "
SELECT transfers.*, 0 as reversed FROM transfers
  JOIN account_relationships  ON  (transfers.id_account_from = account_relationships.child_id)
    WHERE account_relationships.parent_id = \'#{id}\' #{additional_where_sql}
UNION ALL
SELECT transfers.*, 1 as reversed  FROM transfers
  JOIN account_relationships  ON  (transfers.id_account_to = account_relationships.child_id)
    WHERE account_relationships.parent_id = \'#{id}\' #{additional_where_sql}
    #{transfers_sql_order}" },
           :counter_sql => proc { "
SELECT COUNT(*) FROM (
  SELECT transfers.*, 0 as reversed FROM transfers
    JOIN account_relationships  ON  (transfers.id_account_from = account_relationships.child_id)
      WHERE account_relationships.parent_id = \'#{id}\' #{additional_where_sql}
  UNION ALL
  SELECT transfers.*, 1 as reversed  FROM transfers
    JOIN account_relationships  ON  (transfers.id_account_to = account_relationships.child_id)
      WHERE account_relationships.parent_id = \'#{id}\' #{additional_where_sql}
      #{transfers_sql_order}
) as subselect" }

# BASTL from bank account
  has_many :all_transfers_by_datum_seq, :class_name => "Transfer",
           :finder_sql => proc { "
SELECT transfers.*, 0 as reversed FROM transfers
  JOIN account_relationships  ON  (transfers.id_account_from = account_relationships.child_id)
    WHERE account_relationships.parent_id = \'#{id}\' #{additional_where_sql}
UNION ALL
SELECT transfers.*, 1 as reversed  FROM transfers
  JOIN account_relationships  ON  (transfers.id_account_to = account_relationships.child_id)
    WHERE account_relationships.parent_id = \'#{id}\' #{additional_where_sql}
    ORDER BY datum,seq" },
           :counter_sql => proc { "
SELECT COUNT(*) FROM (
  SELECT transfers.*, 0 as reversed FROM transfers
    JOIN account_relationships  ON  (transfers.id_account_from = account_relationships.child_id)
      WHERE account_relationships.parent_id = \'#{id}\' #{additional_where_sql}
  UNION ALL
  SELECT transfers.*, 1 as reversed  FROM transfers
    JOIN account_relationships  ON  (transfers.id_account_to = account_relationships.child_id)
      WHERE account_relationships.parent_id = \'#{id}\' #{additional_where_sql}
      ORDER BY datum,seq
) as subselect" }


  has_many :outgoing_transfers_sorted_by_datum2, :class_name => 'Transfer',
           :finder_sql => proc {"
SELECT transfers.*, 0 as reversed FROM transfers 
  JOIN account_relationships  ON  (transfers.id_account_from = account_relationships.child_id)
    WHERE account_relationships.parent_id = \'#{id}\'
    ORDER BY datum2, date_inserted" }

  has_many :owned_transfers,  :class_name => "Transfer", :foreign_key => "id_account_owner_"
  has_many :documents
  has_many :budgets, :class_name => 'AccountBudget', :foreign_key  => 'account_id'
  has_many :supaccount_relationships, :class_name => 'AccountRelationship', 
        :foreign_key=>'child_id', :order => 'position'
  has_many :supaccounts,
      :through => :supaccount_relationships, :source=>'parent'
  has_many :subaccount_relationships, :class_name => 'AccountRelationship', 
        :foreign_key=>'parent_id', :order => 'position'
  has_many :subaccounts_or_self,
      :through => :subaccount_relationships, :source=>'child'
  has_many :proper_splits, :class_name => 'Split', :foreign_key  => 'account_id'
  has_many :splits, :through=>:subaccounts_or_self, :source=>:proper_splits
  has_many :transfers3, :class_name => 'Transfer', :through=>:splits, :source=>:transfer
  has_many :proper_transfers, :class_name => 'Transfer', :through=>:proper_splits, :source=>:transfer
  
#   has_many :outgoing_transfers2,  # :class_name => "Transfer", :foreign_key => :id_account_from
#       :through => :subaccounts, :source
  has_many :incoming_transfers2,  :class_name => "Transfer", :foreign_key => :id_account_to
  has_many :aliases
  
#   has_and_belongs_to_many :subaccounts,
#         :class_name => 'Account',
#         :join_table => 'account_relationships',
#         :foreign_key => 'parent_id',
#         :association_foreign_key => 'child_id',
#         :order => 'position'

  after_save :generate_self_subaccount_relation
  validates_presence_of :popis
  default_scope order('accounts.seq, accounts.popis')
  
  def all_transfers_sorted_by(column,reversed=false)
    @override_transfers_sql_order=column
    all_transfers
  end

  def transfers_sql_order
    if @override_transfers_sql_order
      temp = @override_transfers_sql_order
      @override_transfers_sql_order = nil  # pro jistotu zapomeneme tridici sloupce
      "ORDER BY #{temp}"
    elsif (id == '3644c25c-0195-11dd-9928-0013d4bc1a7d') or (id == 'c0e3259e-0194-11dd-a278-0013d4bc1a7d') # or (id == '46a0699f-e9db-4cd8-bc50-623e53111b98')  # Daniny, Anaeliny a Jakubovy radime nejdrive podle seq
      'ORDER BY seq,datum2,id'
    elsif prefix_order?
      'ORDER BY LEFT(popis,LOCATE(\':\',popis)-1),datum'
    elsif (id=='ada9b615-dfa5-417e-bc8c-f3d2ce8bc721') or (id=='8b52b9f8-0640-11e1-abd9-7071bcccc8a2') or (id=='46a0699f-e9db-4cd8-bc50-623e53111b98') or 
          (id=='f8a745bc-c2de-11dd-b300-0013d4bc1a7d') or (id=='ada9b615-dfa5-417e-bc8c-f3d2ce8bc721') or (id=='94d1fb1e-ef2c-4cd9-9da7-408034018bc8') or # envelopes
          (id=='b77e749c-b55d-467e-b194-d779a911cdae')  or # Daniel envelope
          (id=='75701c88-d585-11dc-949c-0013d4bc1a7d') # bank cash
      'ORDER BY datum2,seq,id'
    else
      'ORDER BY datum,seq'
    end.
    gsub(/seq/,'IF(reversed=1,IF(LOCATE(\'/\',seq)= 0, seq, RIGHT(seq,LOCATE(\'/\',seq)-1)),seq)')

  end
  
  def additional_where_sql
    if $user=='petra' or $user=='tonda'
      " AND datum>'2009-12-31' "
    elsif $user=='gorn'  # BASTL - testing back in time feature
      '' # " AND date_inserted<'2013-04-12'" # '' # 
    else
      ''
    end
  end

  def inflow_transfers(options={})   # transfery, incoming kladne a outgoing zaporne
    rok = options[:rok].andand.to_i    
    incoming_transfers.select {|t| t.amount.positive? && ( rok.nil? || t.hospodarsky_rok==rok ) } +    
    outgoing_transfers.select {|t| t.amount.negative? && ( rok.nil? || t.hospodarsky_rok==rok ) }    
  end
  def outflow_transfers(options={})    
    rok = options[:rok].andand.to_i    
    incoming_transfers.select {|t| t.amount.negative? && ( rok.nil? || t.hospodarsky_rok==rok ) } +    
    outgoing_transfers.select {|t| t.amount.positive? && ( rok.nil? || t.hospodarsky_rok==rok )}    
  end
  def transit_transfers
    incoming_transfers.select {|t| !t.amount.positive? && !t.amount.negative? } +  
    outgoing_transfers.select {|t| !t.amount.positive? && !t.amount.negative? }        
  end

  def prefix_order?  # jestli se ma radit podle casti popisu pred dvojteckou
    id == 'e50e62c0-dc64-11dd-a205-0013d4bc1a7d' or  # NCSU - personal
    id == '411ada4c-55a2-11de-9b15-0013d4bc1a7d'     # test cash
  end

  def self.all
    if $user == 'mikulas'
      (
        find(OWNER_PALUBA_ID).subaccounts_or_self + 
        find(OWNER_CTYRPOHAR_ID).subaccounts_or_self +
        conditions("type = 'Owner'")
      ).uniq.sort{ |a,b|  (a.is_owner_acct? ? 'b':'a') + a.seq.to_s <=> (b.is_owner_acct? ? 'b':'a') + b.seq.to_s }
    elsif  ['felicia','dana','anael','maria'].include? $user
      hidden = find('fb1062f6-79c4-11dd-ae26-0013d4bc1a7d').andand.subaccounts_or_self +
        find('37367eb8-0734-11dd-90d6-0013d4bc1a7d').andand.subaccounts_or_self

      find(OWNER_NCSU_ID).subaccounts_or_self.select{ |a| not(hidden.include?(a)) }
    elsif $user == 'gorn' or $user == 'jakub'
      super
    elsif $user == 'petra' or $user=='tonda'
      find(OWNER_PALUBA_ID).subaccounts_or_self.sort{ |a,b|  a.seq.to_s <=> b.seq.to_s }
#     elsif $user == 'bisof'
#       find('fe93078e-c239-11df-959a-0013d4bc1a7d').subaccounts_or_self.sort{ |a,b|  a.seq.to_s <=> b.seq.to_s }
    else
      where("accounts.id NOT IN ('e49c00ca-2028-11dd-aa91-0013d4bc1a7d' ,'f924fbee-46ad-11de-85cd-0013d4bc1a7d' ,'b706a06e-21fb-11dd-8f05-0013d4bc1a7d' ,'9ebff020-1c81-11dd-9a3f-0013d4bc1a7d' ,'295fa40a-22db-11dd-8f05-0013d4bc1a7d' ,'e42ddae2-21fd-11dd-8f05-0013d4bc1a7d' ,'a92f5f6c-3a23-11de-9bd3-0013d4bc1a7d' ,'c948eaaa-2202-11dd-8f05-0013d4bc1a7d' ,'b14f9de0-45bb-11dd-8b98-0013d4bc1a7d' ,'6257601c-2200-11dd-8f05-0013d4bc1a7d' ,'1f34d26a-202f-11dd-aa91-0013d4bc1a7d' ,'2d7f7ed8-2200-11dd-8f05-0013d4bc1a7d' ,'cef8ea7c-2685-11dd-b7e2-0013d4bc1a7d' ,'79f71fb0-2033-11dd-aa91-0013d4bc1a7d' ,'2ff76194-21fb-11dd-8f05-0013d4bc1a7d' ,'c5ca5c6e-1c81-11dd-9a3f-0013d4bc1a7d' ,'137a9bd6-31c7-11dd-9145-0013d4bc1a7d' ,'3a327918-dec5-11de-97ab-0013d4bc1a7d' ,'9ea29c0c-04c3-11de-861e-0013d4bc1a7d' ,'9876de2a-fc20-11dd-ad0d-0013d4bc1a7d', '0dc31fe1-654e-45b6-b5e2-891dba0dffb9' )")
    end
  end

  def self.all_assets
    all.select { |account| account.seq =~ /^1|x|^442|^4411|^6/ }
  end

  def self.all_liabilities
    all.select { |account| account.seq =~ /^[0-9]*l/ }
  end

  def self.all_expenses
    all.select { |account| account.expense? }
  end

  def self.all_incomes_and_expenses
    all.select { |account| account.expense_or_income? }
  end

  def self.all_us_expenses
    self.where("seq RLIKE '^[0-9]*u'").order('seq')
  end

  def self.all_owners
    if $user=='petra' or $user=='tonda'
      [Owner.find(OWNER_PALUBA_ID)]
    else
      Owner.order('seq')
    end
  end

  def self.with_undefined_owner
    self.all.select{ |account|  not(account.owner?) and (account.owner_accounts.size == 0) }
  end

  def self.all_without_parent
    self.all.select{ |account| account.supaccounts.empty? }
  end

  # pridava my_account
  def transfers
    outgoing_transfers +
    incoming_transfers
  end

  def rich_outgoing_transfers
    outgoing_transfers
  end

  ### SCOPES and like ###########################################################

  # zda se jedná o účet náležející nějaké pokladně dne účetního kódu
  def pokladna?
    return ((seqofic =~ /^211/) || (seqofic=~ /^365/)) # BASTL PCard NCSU kvuli exportum
  end
  def asset?
    return seq =~ /^1.*/
  end
  def self.where_asset
    where{seq.op('REGEXP', '^1.*')}
  end

  def expense_or_income?
    return seq =~ /^(4|5).*/
  end
  def expense?
    return seq =~ /^4.*/
  end
  def self.where_expense
    where{seq.op('REGEXP', '^4.*')}
  end
  def income?
    return seq =~ /^5.*/
  end
  def type_to_s
    [(self.owner?) ? 'owner' : nil,  (self.asset?) ? 'asset' : nil, (self.expense?) ? 'expense' : nil,
     (self.income?) ? 'income' : nil, (self.debt?) ? 'debt' : nil ].compact.join(', ')
  end
  def self.where_income
    where{seq.op('REGEXP', '^5.*')}
  end

  def debt?
    return seq =~ /^6.*/
  end
  def self.where_debt
    where{seq.op('REGEXP', '^6.*')}
  end

  # rekne, zda se to neda vice specifikovat, cili jestli ma nejake poducty
  def most_specific?
    return (subaccounts_or_self.size==1 and subaccounts_or_self.first == self)
  end
  def is_owner_acct?
    type == 'Owner'
  end
  def is_owner_acct
    type == 'Owner' ? 1 : 0
  end
  def self.where_owner
    where{type=='Owner'}
  end
  def owner_accounts
    supaccounts.where_owner
  end

  # VIRTUAL ATTRIBUTES and OTHER #####################################################

  def pokladna_kod(currency=$default_commodity)
    if (currency==$default_commodity) and (seq =~ /\[(\w*):?(.*)\]/)
      $1
    else
      property[currency]
    end
  end

  def property # pole s všemi vlastnostmi, které se jinam nevešly
    if seq =~ /(\{.*\})/
      YAML::load($1)
    else
      Hash.new(nil) # Hash.new()
    end
  end

  def predkontace
    if seq =~ /\[(.*)\]/  # zatim stejne/podobne jako kod pokladny
      return $1
    else
      return nil
    end
  end

  def owner_account
    owner_accounts.first
  end
  
  def owned_by?(owner)
    owner_accounts.include?(owner)
  end  

  def self.to_accounts_for_owner_id(owner_id)   # BAST asi patri jinam
    if owner_id.nil? or owner_id.empty?
      Account.all_assets.sort{|x,y| x.seq <=> y.seq }
    else
      Account.find(owner_id).subaccounts.select{|a| a.expense_or_income?}.sort{|x,y| x.seq <=> y.seq } +  
      Account.find(owner_id).subaccounts.select{|a| not(a.expense_or_income?) }.sort{|x,y| x.seq <=> y.seq }  + 
      Account.all_assets.sort{|x,y| x.seq <=> y.seq }
    end
  end
	
	def 

  def uses_datum2
    return id=='e49c00ca-2028-11dd-aa91-0013d4bc1a7d'
  end

  def envelope?
    return !responsible_person.nil?
  end
  
  def responsible_person
    responibilities = {
      'c0e3259e-0194-11dd-a278-0013d4bc1a7d'=> "Dana Bartelt", 
      'f8a745bc-c2de-11dd-b300-0013d4bc1a7d'=> "Anna Sedova", 
      'f812efda-93ae-4e56-b607-a58d8c7063b7'=> "Michal Svěrák", 
      '37367eb8-0734-11dd-90d6-0013d4bc1a7d'=> "Jakub Těšínský", # ostatní envelope
      'd9c0c18d-83a8-4c69-a5aa-b39ff14a18a4'=> "Jakub Těšínský", # JT BC
      'b77e749c-b55d-467e-b194-d779a911cdae'=> "Daniel Rektořík", 
      '8b52b9f8-0640-11e1-abd9-7071bcccc8a2'=> "Zuzana Černohousová", 
      '9ed801e7-a479-4917-a76d-580fbbfd067f'=> "Zuzana Černohousová", 
      '3644c25c-0195-11dd-9928-0013d4bc1a7d'=> "Anael Segers", 
      'c4a1c508-2d1d-11df-9784-0013d4bc1a7d'=> "Anael Segers", 
      '560236aa-b9d1-11df-90bf-0013d4bc1a7d'=> "Maria Sokolovská",
      '541b745e-016a-11dd-8e74-0013d4bc1a7d'=> "Prague Institute faculty",
      '8a344e1a-11aa-11df-9f90-0013d4bc1a7d'=> "Anael Segers",
      '2acae616-016a-11dd-ae79-0013d4bc1a7d'=> "Prague Institute faculty",
      '62e3b9c6-d585-11dc-b2dd-0013d4bc1a7d'=> "Dana Bartelt", 
      'b1dadff7-dff0-4945-83ae-9bab49f29eaf'=> "Peter Kjær",
      '1b9c2975-69e2-4685-902c-dada1ae2b80d'=> "Peter Kjær",
      '74729926-3151-47d4-8238-eca98f9ac097'=> "Peter Kjær"
    }
    responibilities[id]
  end
  
  def recursive_subaccounts(uptolevel=4)
    result = self.subaccounts
    if uptolevel>1
      for account in self.subaccounts
        result += account.recursive_subaccounts(uptolevel-1)
      end
    end
    result.uniq!
    return result
  end

  def subaccounts
    subaccounts_or_self - [self]
  end
  
  def add_subaccount(account)
    subaccounts_or_self << account unless subaccounts_or_self.include?(account)
  end
  
  def subaccounts=(value)
    subaccounts_or_self= value + [self]
  end
  
  def owner?
    type == 'Owner'
  end 

  def related_owners
    not_defined
  end

  def owners
    all_transfers.collect{ |t| t.owner }.uniq
  end

  def related_accounts
    all_transfers.collect{ |t| t.id_other_account }.uniq.collect {|id| Account.find(id)}.reject{|a| a.nil?}
  end

  def balance_from_account(account)
    - balance_to_account(account)
  end
  
  def balance_to_account(foreign_account, options = {})
    subaccounts_of_foreign = foreign_account.subaccounts_or_self
    sum = {}
    rok = options[:rok].andand.to_i    
    all_transfers.each{ |t|
      if subaccounts_of_foreign.include?(t.other_account) && ( rok.nil? || t.hospodarsky_rok==rok )
        sum += t.my_amount
      end
    }
    sum
  end

  def balance_from_account_without_subaccounts(account)
    - balance_to_account_without_subaccounts(account)
  end

  def balance_to_account_without_subaccounts(foreign_account)
    sum = {}
    all_transfers.each{ |t|
      if foreign_account == t.other_account
        sum += t.my_amount
      end
    }
    sum
  end

  def balance_to_account_by_owners_without_subaccounts(foreign_account)
    result = {}
    all_transfers.each{ |t|
      if (foreign_account == t.other_account)
        if result[t.owner.id].nil?
          result[t.owner.id] = t.my_amount
        else 
          result[t.owner.id] += t.my_amount
        end
      end
    }
    result
  end

  def balance_for_vendor(vendor)
    all_transfers.select{ |t| t.vendor == vendor }.inject({}){|s,t| s += t.my_amount}
  end

  def balance(options = {})
    if options[:epoch]
      epoch = options[:epoch]
      all_transfers.inject({}){|s,t| 
        s += t.my_amount  if t.epoch_id==epoch.id
        s
      }
    else # stara implementace primo s rokem
      rok = options[:rok].andand.to_i    
      if rok.nil?
        all_transfers.inject({}){|s,t| s += t.my_amount}
      else
        all_transfers.inject({}){|s,t| 
          s += t.my_amount  if t.hospodarsky_rok==rok
          s
        }
      end
    end
  end

  def inflow_balance(options={})
    - inflow_transfers(options).inject({}){|s,t| s += t.my_amount }
  end
  def outflow_balance(options={})
    - outflow_transfers(options).inject({}){|s,t| s += t.my_amount }
  end

  def balance_using_outgoing_transfers
    outgoing_transfers.inject({}){ |s,t| s += t.amount}
    
#     balance = {}
#     self.outgoing_transfers_sorted_by_datum2.reject! do |t| 
#       balance += t.my_balance
#     end
#     balance
    
#     subaccounts = self.subaccounts_or_self.collect{ |a| a.id }
#     outgoing_transfers.delete{ |t|
#       (subaccounts.include?(t.account_from.id) and subaccounts.include?(t.account_to.id)
#     }.inject({}){ |s,t| s += t.amount}
  end

  def owner_balance(owner)
    - all_transfers.select{ |t| t.owner == owner }.inject({}){|s,t| s += t.my_amount}
  end

  def balance_to_date(date)
    all_transfers.select{ |t| t.datum <= date }.inject({}){|s,t| s += t.my_amount}
  end

  def balance_in_hospodarsky_rok(hospodarsky_rok)
    all_transfers.select{ |t| (hospodarsky_rok.start_date <= t.datum) && (t.datum <= hospodarsky_rok.end_date) }.inject({}){|s,t| s += t.my_amount} 
  end

  def asset_subaccounts
    subaccounts_or_self.where_asset
  end

  def expense_subaccounts
    subaccounts_or_self.where_expense
  end

  def debt_subaccounts
    subaccounts_or_self.where_debt
  end

  def income_subaccounts
    subaccounts_or_self.where_income
  end

  def immediate_subaccounts
    subaccounts - subaccounts.collect{ |sa| sa.subaccounts }.flatten
  end

  def parents # alias immediate_supaccount
    supaccounts - supaccounts.collect{ |sa| sa.supaccounts }.flatten
  end

  def rodokmen # vraci strom pustupnych rodicu. [ [rodice], [prarodice], ... ]
    current = [self]
    result = []
    while !current.empty?
      result << current
      current = current.collect{|a| a.parents}.flatten.uniq
    end
    result
  end

  def self.rodokmen_of_array(accounts)
    
  end

  def immediate_subaccount_relationships
    subaccount_relationships - subaccounts.collect{ |sa| sa.subaccount_relationships }.flatten
  end

  def last_seq
    all_transfers.max {|t1,t2| t1.my_seq.to_i<=>t2.my_seq.to_i }.seq.to_i
  end

  def next_seq
   last_seq + 1
  end

  def next_seq2(year)  # vrátí první neobsazené seq2 v daném roce v pokladnách se stejnou zkratkou
    @pokladna_kod = pokladna_kod.to_s  # kod pokladny ktere se to tyka
    @pokladna_prefix =  @pokladna_kod.gsub(/\d\d\d$/,'') # prefix kodu jednotlivych transferu (momentalne sezere tri posledni cislice)
    if !@pokladna_kod.blank?
      @posledni_transfer =  Account.all.select{ |a| a.pokladna_kod == @pokladna_kod }.collect{ |a| a.transfers }.flatten.select{ |t|
        if !t.seq2.nil?
          if year =~ /^\d*$/
            (t.datum.year == year)
          else
              fromyear = year.split('-')[0]
              toyear = year.split('-')[1]
              ((t.datum>= Date.civil(fromyear.to_i,7,1)) and (t.datum<= Date.civil(toyear.to_i,6,30)))
          end
        else
          false
        end
      }.max { |a,b|
        a.seq2.gsub(/^#{@pokladna_prefix}/,'').to_i <=> b.seq2.gsub(/^#{@pokladna_prefix}/,'').to_i
      }
    end

    if @posledni_transfer.nil?
      @poradi = 1
    else
      @poradi = @posledni_transfer.seq2.gsub(/^#{@pokladna_prefix}/,'').to_i + 1
    end
		
    return @pokladna_prefix + ('%03d' % @poradi.to_s)
  end
  
  def self.unidentified
    Account.find(UNIDENTIFIED_ACCOUNT_ID)
  end

  def all_used_years  # vrati seznam pouzitych roku
    transfers=self.all_transfers
    if transfers.empty?
      []
    else
      miny, maxy = transfers.inject([transfers.first.datum,transfers.first.datum]){ |res,t| 
        [ [res[0],t.datum].min, [res[1],t.datum].max ]      
      }      
      miny.year.upto(maxy.year).collect
    end
  end
  def minimal_date
    minmaxdates.first
  end
  def maximal_date
    minmaxdates.last
  end
  def is_owner_acct?
    type == 'Owner'
  end
  def is_owner_acct
    type == 'Owner' ? 1 : 0
  end
  def is_owner_acct=(value)
    if value=='1' or value == 1
      self.type='Owner'
    else
      self.type='Account'
    end
  end
  
  has_many :new_outgoing_transfers, :class_name => "Transfer", :foreign_key => "id_account_from"
  has_many :new_incoming_transfers, :class_name => "Transfer", :foreign_key => "id_account_from"

  def new_transfers
    myid = self.id
    Transfer.joins('INNER JOIN reverse').select('transfers.*, reverse.*').where{(id_account_from == myid) && (reverse.reversed == 0)}
  end
  # vrátí závěrečnou část popisu kteru se liší od svého rodičovského účtu - musí být odděleno lomítkem
  def popis_suffix
    p = parents
    if (p.size==1) && (p[0].popis+"#"+popis =~ /^(.*).*#\1\s*\/\s*(.*)/)
      $2
    else
      popis
    end
  end
  # ----------------------------------------------
  def balance2(options = {})
    s = amounts_string(options)
    amount_string2amount(s)
  end
  def amounts_strings
    ActiveRecord::Base.connection.execute(
      transfers_with.
        select("transfers.id_account_to, transfers.epoch_id, GROUP_CONCAT(amount SEPARATOR ';') as amount_string").
        joins{account_from.supaccounts}.
        group('transfers.id_account_to, transfers.epoch_id').to_sql
    )
  end  
  def transfers_with(options = {})
    if options[:epoch]
      all_transfers2.merge(options[:epoch].transfers.scoped)
    else
      all_transfers2
    end
  end
  def all_transfers2
    outgoing_transfers2 #.union(incoming_transfers2)
  end
  def outgoing_transfers2
    aid = self.id
    Transfer.select('transfers.*, 0 as reversed').joins{account_from.supaccounts}.
      where{(account_from.supaccounts.id==aid) | (account_from.id==aid)}
  end
  def incoming_transfers2
    aid = self.id
    Transfer.select('transfers.*, 1 as reversed').joins{account_to.supaccounts}.
      where{(account_to.supaccounts.id==aid) | (account_to.id==aid)}
  end
  def filter_reconcilable(transfers, options = {}) ## this is used in us_report. it finds first unreconciled transfer and takes all Transfer after it. It ignores all circular transfers and calculated starting balance
    balance = {}
    if options[:include_reconciled] 
      still_header = false # pokud je options[:include_reconciled] zvoleno, tak chci vsechny a rovnou od zacatku nejsem v hlavicce
    else
      still_header = true
    end
    something_was_already_accepted = false # jen pro ucely warningu
    subaccounts = self.subaccounts_or_self.collect{ |a| a.id }
    hide_michailed = (self.id != '9a9726a4-f5b6-11dc-85fa-0013d4bc1a7d')

    result_transfers = transfers.reject do |t| # count starting balance and hide starting reconciled/michailed transfers
      if (still_header and t.reconciled?) or (hide_michailed and t.michailed) or
         (subaccounts.include?(t.account_from.id) and subaccounts.include?(t.account_to.id))
        
        unless (subaccounts.include?(t.account_from.id) and subaccounts.include?(t.account_to.id)) # circular transfer
          balance += t.my_balance  # TODO: u circular transferu by prece mohlo ale byt t.my_balance rovno 0 ne?
        end
#        flash[:warning] = "Transfer #{t.id} is Med so balance will jump."  if something_was_already_accepted
        true  # return true to reject! the transfer
      else
        still_header = false
        something_was_already_accepted = true
        false # return false to accept! the transfer
      end
    end
   
#    if $user=='jakub' 
#       result_transfers = transfers.reject do |t|      
#         false
#       end
#        
#      raise result_transfers.inspect 
#    end

    return result_transfers, balance  
  end
  def amount_string2amount(amount_str)
    if amount_str.nil?
      {}
    else
      begin
        amount_str.split(';').collect{|amount| (/^---/.match(amount) ? YAML::load(amount.chomp!) : amount).to_multicash_h }.sum
      rescue
        raise "ERROR: #{amount_str} ///\n#{amount_str.gsub(/;?---/,"\n-")}"
      end
    end
  end  
  def amounts_string(options = {})   # ladici mezifunkce
    ActiveRecord::Base.connection.execute(
      transfers_with(options).select("GROUP_CONCAT(amount SEPARATOR ';')").to_sql
    ).first.first
  end
  def balance2_table
    res = {}
    amounts_strings.each(:as => :hash, :symbolize_keys => true) do |row|
      (res[row[:id_account_to]] ||= {})[row[:epoch_id]]=amount_string2amount(row[:amount_string])
    end
    res
  end
  def to_s
    popis.to_s
  end  
  
  # splits_with(options) returns splits which satisfy some options (returns scope) 
  # for options see Split.all_with
  # vrácený scope má najoinováno accounts
  def splits_with(options={})
    Split.all_with(options.merge({:account=>self}))
  end
  def balance3(options={})
    splits_with(options).joins{amounts}.group{amounts.currency}.sum(:value)
  end
  def balance3_table
    res = {}
#     splits_with({}).joins{amounts}.joins{transfer}.group{[account_id,transfers.epoch_id,amounts.currency]}.sum(:value)
    splits_with({}).joins{amounts}.joins{transfer}.select('sum(`amounts`.`value`) AS av, `splits`.`account_id` AS sai, `transfers`.`epoch_id` AS tei').group('sai, tei, amounts.currency').all.each do |key|
      account_id = key[0]
      epoch_id = key[1]
      currency = key[2]
      ((res[account_id] ||= {})[epoch_id] ||={})[currency]=key[2]
    end
    res
  end
  def unreconciled_splits          # nereconcilované splity (tj, které mají nereconcilovaný transfer)
    splits.joins{transfer}.where{ |s| s.transfer.reconciled != true }
  end  
  def unreconciled_transfers_count # kolik transferů v accaountu není reconcilováno
    unreconciled_splits.size
  end
  def generate_self_subaccount_relation
    self.add_subaccount(self) unless self.subaccounts_or_self.include?(self)
  end
  def popis_without(prefix)
    if !prefix.is_a?(Array)
      prefix=[prefix]
    end
    popis.gsub(/^(#{prefix.compact.join('|')})\s*\/?\s*/,'')
  end
  # udela stromove pole s mnoziny accountu. jedna polozka pole vypada takto
  # { :indexes => [0,1], :account=><account object>, :has_children=><false/true> }
  # predpokladam ze jsou accounty usporadany v "hierarchickem poradi"
  # @param accs [Array] pole hierarchicky uspořádaných účtů
  def self.construct_tree_array(accs)
    accs=accs.all if !accs.is_a? Array 
    return [] if accs.last==nil # tohle hazi Cannot visit Arel::Nodes::InfixOperation
    startindex = 0
    tree = [{:indexes => [startindex], :account=>accs[0], :has_children=>false}]
    path = tree.clone
    level = 1
    accs[1..-1].each do |account|
      current = {:account=>account, :has_children=>false}
      sibbling = nil
      # odzadu hledej v path naducet currentu
      until path.empty? or path.last[:account].subaccounts_or_self.include?(account)
        sibbling = path.pop
      end
      
      path.last[:has_children]=true unless path.empty?
      if sibbling.nil?
        # zanořit se do podůčtu
        current[:indexes]= path.last[:indexes].clone + [startindex]
        current[:parent]= path.last[:account]
      else
        current[:indexes]= sibbling[:indexes].clone
        current[:indexes][-1] += 1
        current[:parent]= sibbling[:parent]
      end
      path.push(current)
      tree.push(current)
    end
    tree      
  end
end